/* 
   Copyright (c) 2001 Paul Campbell - paul@taniwha.com
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/
                           
#include <stdio.h>
#include <dirent.h>
#include "resources.h"

//
//	makeresource root -o output -c comment
//

char *prog;
int nocvs;
FILE *fout;

struct entity {
	res_u64		size;
	res_u64		offset;
	res_u64		base;
	mode_t		mode;
	int		inode;
	struct entity	*next;	
	struct entity	*prev;	
	struct entity	*child;	
	struct entity	*parent;	
	char		*name;
	char		*hier_name;
	struct res_inode *ip;
};
int tsize=0;		// sum of sizes of names (null terminated and rounded up to 4)
int ninodes=0;		// number of total inodes
int ndirs=0;		// number of directories
int filesize=0;		// total file sizes in bytes (each rounded up to multiple of 4 bytes
int verbose=0;
int inode = 2;
char *output=NULL;

void
no_space()
{
	fprintf(stderr, "%s: out of virtual memory - buy DRAM futures\n", prog);
	exit(1);
}


struct entity *
build_hier(char *file, char *name, struct entity *parent)
{
	struct stat s;
	struct entity *ep;

	if (stat(file, &s)) {
		fprintf(stderr, "%s: can't stat file %s\n", prog, file);
		exit(1);
	}
	if (S_ISLNK(s.st_mode)) {
		fprintf(stderr, "%s: warning symbolic links are not supported %s\n", prog, file);
		return(NULL);
	}
	if (S_ISCHR(s.st_mode) || S_ISBLK(s.st_mode)) {
		fprintf(stderr, "%s: warning special files are not supported %s\n", prog, file);
		return(NULL);
	}
	if (S_ISFIFO(s.st_mode) || S_ISSOCK(s.st_mode)) {
		fprintf(stderr, "%s: warning named pipes/fifos are not supported %s\n", prog, file);
		return(NULL);
	}
	if (S_ISREG(s.st_mode)) {
		ep = (struct entity *)malloc(sizeof(*ep));
		if (ep == NULL)
			no_space();
		filesize += (s.st_size+3)&~3;
		tsize += (strlen(name)+1+3)&~3;
		ninodes++;
		ep->inode = inode++;
		ep->size = s.st_size;
		ep->name = name;
		ep->mode = s.st_mode;
		ep->hier_name = file;
		ep->parent = parent;
		ep->next = NULL;
		ep->prev = NULL;
		ep->child = NULL;
		return(ep);
	} else
	if (S_ISDIR(s.st_mode)) {
		DIR *dp;
		struct dirent *entp;
		struct entity *ep2, *ep3;
		char *cp1, *cp2;

		ep = (struct entity *)malloc(sizeof(*ep));
		if (ep == NULL)
			no_space();
		tsize += (strlen(name)+1+3)&~3; 
		ninodes++;
		ndirs++;
		ep->size = 0;	// number of inodes
		ep->inode = inode++;
		ep->child = NULL;
		ep->prev = NULL;
		ep->next = NULL;
		ep->mode = s.st_mode;
		ep->hier_name = file;
		ep->name = name;
		dp = opendir(file);
		if (dp == NULL) {
			fprintf(stderr, "%s: can't read directory - %s - ignored\n", prog, file);
			return(NULL);
		} 
		while ((entp = readdir(dp)) != NULL) {
			if (strcmp(entp->d_name, ".") == 0)
				continue;
			if (strcmp(entp->d_name, "..") == 0)
				continue;
			if (nocvs && strcmp(entp->d_name, "CVS") == 0)
				continue;
			cp1 = (char *)malloc(strlen(entp->d_name)+1);
			if (cp1 == NULL)
				no_space();
			strcpy(cp1, entp->d_name);
			cp2 = (char *)malloc(strlen(entp->d_name)+1+strlen(file)+1);
			if (cp2 == NULL)
				no_space();
			strcpy(cp2, file);
			strcat(cp2, "/");
			strcat(cp2, entp->d_name);
			ep2 = build_hier(cp2, cp1, ep);
			if (ep2 == NULL) {
				free(cp1);
				free(cp2);
				continue;
			}
			//
			//	sort the entries into the hierarchy
			//
			ep->size++;
			if (ep->child == NULL) {
				ep->child = ep2;
				ep2->next = NULL;
				ep2->prev = NULL;
			} else {
				for (ep3 = ep->child; ep3; ep3=ep3->next) {
					if (strcmp(ep2->name, ep3->name) < 0) {
						if (ep3->prev == NULL) {
							ep->child = ep2;
						} else {
							ep3->prev->next = ep2;
						}
						ep2->prev = ep2->prev;
						ep3->prev = ep2;
						ep2->next = ep3;
						break;
					}
					if (ep3->next == NULL) {
						ep3->next = ep2;
						ep2->prev = ep3;
						ep2->next = NULL;
						break;
					}
				}
			}
		}
		closedir(dp);
		return(ep);
	} else {
		fprintf(stderr, "%s: unknown type of file %s - ignored\n", prog, file);
		return(NULL);
	}
}

void
dump_hierarchy(int ind, struct entity *ep)
{
	int i;

	for (i = 0; i < ind; i++)
		printf("  ");
	if (S_ISREG(ep->mode)) {
		printf("%s - file size %d mode 0x%x from %s\n", ep->name, (unsigned long)ep->size, ep->mode, ep->hier_name);
	} else {
		printf("%s - dir size %d mode 0x%x\n", ep->name, (unsigned long)ep->size, ep->mode);
		for (ep = ep->child; ep; ep=ep->next)
			dump_hierarchy(ind+1, ep);
	}
}

unsigned char *data;
res_u64 size;

res_u64
build_inodes(struct entity *epp)
{
	struct entity *ep;
	res_u64 base, i;
	struct res_inode *ip;

	if (epp->child == NULL)
		return(0);
	base = 0;
	for (ep = epp->child; ep; ep=ep->next) {	
		if (base == 0)
			base = size;
		ep->base = i = size;
		ep->ip = ip = (struct res_inode *)&data[size];
		ip->size = ep->size;
		ip->nlinks = 1;
		ip->parent = epp->base;
		ip->mode = ep->mode;
		ip->inode = ep->inode;
		ip->pinode = epp->inode;
		if (S_ISDIR(ep->mode)) {	
			epp->ip->nlinks++;	// the .. link
			ip->nlinks++;		// the . link
		}
		strcpy(ip->name, ep->name);
		size += sizeof(*ip)+strlen(ep->name)+1;
		size = (size+3)&~3;
		if (ep->next) {
			ip->next = size-i;
		} else {
			ip->next = 0;	
		}
	}
	for (ep = epp->child; ep; ep=ep->next) {	
		ep->ip->offset = build_inodes(ep);
	}
	return(base);

}

void
copy_files(struct entity *ep)  
{
	static char tmp[40960];
	res_u64 s;
	int l;
	FILE *f;

	if (S_ISREG(ep->mode)) {
		ep->ip->offset = size;
		f = fopen(ep->hier_name, "r");
		if (f == NULL) {
			fprintf(stderr, "%s: can't open file '%s'\n", prog, ep->hier_name);
			exit(1);
		}
		s = 0;
		for (;;) {
			l = fread(tmp, 1, sizeof(tmp), f);
			if (l <= 0)
				break;
			if (fwrite(tmp, 1, l, fout) != l) {
				fprintf(stderr, "%s: error writing output file '%s'\n", prog, output);
				exit(1);
			}
			s += l;
		}
		fclose(f);
		ep->ip->size = s;
		size = (size+s+3)&~3;
	} else {
		for (ep = ep->child; ep; ep=ep->next) 
			copy_files(ep);
	}
}

int
main(int argc, char **argv)
{
	char *comment = "";	
	char *root=NULL;
	int i;
	res_u64 xsize;
	struct entity *ep;
	struct res_hdr *rh;

	prog = argv[0];
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-nocvs") == 0) {
			nocvs = 1;
			continue;
		}
		if (strcmp(argv[i], "-v") == 0) {
			verbose = 1;
			continue;
		}
		if (strcmp(argv[i], "-help") == 0) {
			fprintf(stderr, "Usage: %s [-v][-nocvs] rootdir -o outputfile\n", argv[0]);
			continue;
		}
		if (strcmp(argv[i], "-c") == 0) {
			i++;
			if (i == argc) {
				fprintf(stderr, "%s: missing comment for -c\n", argv[0]);
				exit(1);
			}
			comment = argv[i];
			continue;
		}
		if (strcmp(argv[i], "-o") == 0) {
			i++;
			if (i == argc) {
				fprintf(stderr, "%s: missing output for -o\n", argv[0]);
				exit(1);
			}
			if (output) {
				fprintf(stderr, "%s: only one -o flag allowed\n", argv[0]);
				exit(1);
			}
			output = argv[i];
			continue;
		}
		if (root) {
			fprintf(stderr, "%s: only one root directory allowed\n", argv[0]);
			exit(1);
		}
		root = argv[i];
	}
	if (root == NULL) {
		fprintf(stderr, "%s: no root directory specified\n", argv[0]);
		exit(1);
	}
	if (output == NULL) {
		fprintf(stderr, "%s: no output file (-o) specified\n", argv[0]);
		exit(1);
	}
	ep = build_hier(root, "", NULL);
	if (verbose)
		dump_hierarchy(0, ep);
	data = (unsigned char *)malloc(sizeof(struct res_hdr)+ tsize + ninodes*sizeof(struct res_inode)+
		strlen(comment)+4);
	if (data == NULL)
		no_space();
	size = 0;

	rh = (struct res_hdr *)&data[size];
	size += sizeof(*rh);
	strncpy(rh->version, RES_VERSION_1_0, sizeof(rh->version));
	rh->time = time(NULL);
	strcpy(rh->comment, comment);
	size += strlen(comment)+1;	
	size = (size+3)&~3;
	ep->ip = &rh->root;
	ep->base = (unsigned char *)(&rh->root) - data;
	rh->root.inode = ep->inode;	
	rh->root.pinode = ep->inode;	
	rh->root.parent = (long)(&rh->root) - (long)(rh);	
	rh->root.nlinks = 2;
	rh->root.size = ep->size;
	rh->root.mode = ep->mode;
	rh->root.next = 0;
	rh->root.offset = build_inodes(ep);
	fout = fopen(output, "w");
	if (fout == NULL) {
		fprintf(stderr, "%s: Can't open output file '%s'\n", argv[0], output);
		exit(1);
	}
	xsize = size;
	fseek(fout, size, SEEK_SET);	// skip past the right place
	copy_files(ep);
	rewind(fout);
	if (fwrite(data, xsize, 1, fout) != 1) {
		fprintf(stderr, "%s: Error writing output file '%s'\n", argv[0], output);
		exit(1);
	}
	fclose(fout);
	exit(0);
}
