/* 
   Copyright (c) 2001 Paul Campbell - paul@taniwha.com
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/
                           
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>   
#include <unistd.h>
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include <linux/elf.h>

#include "resources.h"



#define RES_PATH	"/proc/self/resource/" 	// alternate interface

//
//	convert to /proc/self/resource/<filename>
//
static char *
tmp_path(const char *cp)
{
	char *tcp;

	tcp = (char *)malloc(strlen(cp)+1+strlen(RES_PATH));
	if (tcp == NULL)
		return(tcp);
	strcpy(tcp, RES_PATH);
	strcat(tcp, cp);
	return(tcp);
}

//
//	read the inode structure from offset
//
static struct res_inode *
rd_ip(struct RES_ENV *rf, res_u64 offset)
{

	if (fseek(rf->f, rf->total_base+offset, SEEK_SET))
		return(NULL);
	if (fread(rf->ip, 1, sizeof(rf->ip), rf->f) < (sizeof(struct res_inode)+2))
		return(NULL);
	return((struct res_inode *)rf->ip);
}

//
//	open a resource file - here's where the code file specific stuff goes
//
struct RES_ENV *
res_open_env(const char *name)
{
	struct RES_ENV *rf;
	struct elf32_hdr ehdr;
	Elf32_Shdr shdr; 
	long l, i;

	rf = (struct RES_ENV *)malloc(sizeof(*rf));
	if (rf == NULL) {
		return(NULL);
	}
	rf->have_res = 0;	
	if (name == NULL) {
		if (!access(RES_PATH, R_OK)) {
			rf->have_res = 1;
			rf->f = NULL;	
			rf->ref_count = 1;
			return(rf);
		}
	}
	rf->f = fopen("/proc/self/exe", "r");				// open open it
	if (rf->f == NULL) {
		free(rf);
		return(NULL);						// can't? fail
	}

	//
	//	now try and get data from an ELF file
	//
	l = fread(&ehdr, 1, sizeof(ehdr), rf->f);
	if (l != sizeof(ehdr)) 
		goto try_file2;
	if (ehdr.e_ident[0] != 0x7f ||
            strncmp(&ehdr.e_ident[1], "ELF", 3) != 0)
               	goto try_file2; 
	for (i = ehdr.e_shnum-1; i >= 0; i--) {				// go down because it's near the end
		if (fseek(rf->f, ehdr.e_shoff+i*ehdr.e_shentsize, SEEK_SET) != 0)
			goto try_file2;
		if (fread(&shdr, 1, ehdr.e_shentsize, rf->f) != ehdr.e_shentsize)
			goto try_file2;
		if (shdr.sh_type != SHT_PROGBITS)
			continue;
		if (shdr.sh_flags != SHF_WRITE)
			continue;
		if (shdr.sh_size == 0)
			continue;
		if (fseek(rf->f, shdr.sh_offset, SEEK_SET)) 
			continue;
		//
		// 	at this point it's the right type - check the header
		//	note - there's no point in checking the sections name,
		//	it's cheaper to just read the header and check it's
		//	version
		//
		if (fread(&rf->hdr, 1, sizeof(rf->hdr), rf->f) != sizeof(rf->hdr))
			continue;
		if (strncmp(RES_VERSION_1_0, &rf->hdr.version, strlen(RES_VERSION_1_0)) == 0) {
			rf->total_base = shdr.sh_offset;	// set the base offset
			rf->total_size = shdr.sh_size;		// and size
			goto ok;
		}
	}
try_file2:
	//
	//	we didn't find an ELF file ... 
	//	here's where we add new object file types 
	//
	fclose(rf->f);
	free(rf);
	return(NULL);
	//
	// success - return the structure
	//
ok:
	rf->ref_count = 1;
	return(rf);
}

void
res_close_env(struct RES_ENV *rf)
{
	rf->ref_count--;
	if (rf->ref_count <= 0) {
		if (rf->f)
			fclose(rf->f);
		free(rf);
	}
}

//
//	find the inode pointed to by <path>, check to see if it's
//		the right kind and return an open descriptor to it
//
static struct RES_FD *
get_fd(struct RES_ENV *rf, const char *path, int want_file, int want_dir)
{
	struct res_inode *ip; 
	res_u64 inode_offset;
	res_u64 inode_parent;
	res_u64 inode_child;
	struct RES_FD *fd;
	const char *cp;
	char *cp2;
	char tmp[NAME_MAX+2];

	rf->ref_count++;					// mark it busy
	ip = &rf->hdr.root;					// pointer to current inode
	inode_offset = ip->parent;				// offset of our inode
	inode_parent = ip->parent;				// offset of parent inode
	inode_child = ip->offset;				// offset of list of children

	for (cp = path;*cp;) {					// loop thru text
		while (*cp == '/')				// skip starting /s
			cp++;
		if (*cp == '.') {			
			cp++;
			if (*cp == '/' || *cp == 0)		// ignore "."
				continue;
			if (*cp == '.' && (cp[1] == '/' || cp[1] == 0)) {	
				cp++;
				ip = rd_ip(rf, inode_parent);	// ".." move to parent
				inode_offset = inode_parent;
				inode_parent = ip->parent;
				inode_child = ip->parent;
				continue;
			}
		}
				
		if (cp == 0) 					// at end we've found something
			break;
		if (inode_child == 0) {				// no children means 
bad:								// we can't find it
			rf->ref_count--;
			rf->location = 0;
			if (rf->ref_count == 0)	{		// fail so let go of the file
				fclose(rf->f);			// close it if we're last
				free(rf);
			}
			return(NULL);
		}
		cp2 = tmp;					// copy the file segment portion
		for (;;) {
			*cp2++ = *cp++;
			if (*cp == '/' || *cp == 0)
				break;
		}
		*cp2 = 0;
		for (;;) {
			ip = rd_ip(rf, inode_child);		// walk thru children comparing it
			if (ip == NULL)
				goto bad;
			if (strcmp(tmp, ip->name) == 0) {	// found it brteak
				break;
			}
			if (ip->next == 0)			// can't find it fail
				goto bad;
			inode_child += ip->next;		// move to next
		}
		inode_offset = inode_child;			// point to next directory
		inode_parent = ip->parent;			// and loop to get next part of the name
		inode_child = ip->offset;
	}
	//
	// 	now check to see if it's what we want
	//
	if (S_ISREG(ip->mode)) {
		if (!want_file)
			goto bad;
	} else
	if (S_ISDIR(ip->mode)) {
		if (!want_dir)
			goto bad;
	} else goto bad;
	//
	//	allocate a descriptor structure
	//	and fill it in
	//
	fd = (struct RES_FD *)malloc(sizeof(*fd));
	if (fd == NULL)
		goto bad;
	fd->rf        = rf;
	fd->base      = ip->offset;	// where it's data is
	fd->size      = ip->size;		// how big
	fd->parent    = ip->parent;	// where it's parent is
	fd->addr      = inode_offset;	// where the inode is
	fd->a1.inode  = ip->inode;
	fd->a2.pinode = ip->pinode;
	fd->a3.mode   = ip->mode;
	fd->offset    = 0;		// current seek
	rf->location  = 0;
	return(fd);			// success!!
}

//
//
struct RES_FD *
res_open(struct RES_ENV *xrf, const char *pathname, int flags)
{
	struct RES_ENV *rf;
	char *cp;
	struct RES_FD *fd;

	if (xrf == NULL) {
		rf = res_open_env(NULL);
		if (rf == NULL)
			return(NULL);
	} else {
		rf = xrf;
	}
	if (rf->have_res) {
		int i;

		cp = tmp_path(pathname);		// create real name
		if (cp == NULL)
			return(NULL);
		i = open(cp, flags);		//  open it
		if (i < 0) {
			res_close_env(rf);
			fd = NULL;
		} else {
			fd = (struct RES_FD *)malloc(sizeof(*fd));
			if (fd == NULL) {
				if (xrf == NULL)
					res_close_env(rf);
			} else {
				fd->rf      = rf;
				fd->a1.fd   = i;
				fd->a2.f    = NULL;
				fd->a3.dp   = NULL;
			}
		}
		free(cp);
		return(fd);
	}
	fd = get_fd(rf, pathname, 1, 0);			// else get a file desc
	if (xrf == NULL)
		res_close_env(rf);
	return(fd);
}

int
res_close(struct RES_FD *fd)
{
	if (fd->rf->have_res) {
		(void)close(fd->a1.fd);		
	}
	res_close_env(fd->rf);
	free(fd);
	return(0);
}

ssize_t
res_read(struct RES_FD *fd, void *buf, size_t count)
{
	res_u64 l;

	if (fd->rf->have_res) return(read(fd->a1.fd, buf, count));
	l = fd->size-fd->offset;				// amount left
	if (count < l)
		l  = count;					// min of req and amount left
	if (fd->rf->location != (fd->rf->total_base+fd->base+fd->offset)) {	// seek if required
		if (fseek(fd->rf->f, fd->rf->total_base+fd->base+fd->offset, SEEK_SET))
			return(-1);
		fd->rf->location = fd->rf->total_base+fd->base+fd->offset;
	}
	l = fread(buf, 1, l, fd->rf->f);			// read the data
	if (l > 0) {
		fd->rf->location += l;					// update seek location
		fd->offset += l;
	}
	return(l);
}

int
res_rstat(struct RES_ENV *xrf, const char *file_name, struct stat *buf)
{
	struct RES_FD *fd;
	struct RES_ENV *rf;
	int i;

	if (xrf == NULL) {
		rf = res_open_env(NULL);
		if (rf == NULL) {
			errno = ENOENT;
			return(-1);
		}
	} else {
		rf = xrf;
	}
	if (rf->have_res) {
		file_name = tmp_path(file_name);
		if (file_name == NULL)
			return(-1);
		i = stat(file_name, buf);
		free(file_name);
		if (xrf == NULL) 
			res_close_env(rf);
		return(i);
	}
	fd = get_fd(rf, file_name, 1, 1);			// open a file
	if (fd == NULL) 
		return(-1);
	i = res_fstat(fd, buf);					// do fstat
	res_close(fd);						// and close it
	if (xrf == NULL)
		res_close_env(rf);
	return(i);
}

int
res_stat(const char *file_name, struct stat *buf)
{
	struct RES_FD *fd;
	int i;

	return(res_rstat(NULL, file_name, buf));
}

int
res_fstat(struct RES_FD *fd, struct stat *buf)
{
	struct res_inode *ip;

	if (fd->rf->have_res) return(fstat(fd->a1.fd, buf));
	ip = rd_ip(fd->rf, fd->addr);
	if (ip == NULL)
		return(-1);
	buf->st_dev = 0;				// fake out the data
	buf->st_ino = ip->inode;
	buf->st_mode = ip->mode;
	buf->st_nlink = ip->nlinks;
	buf->st_uid = getuid();
	buf->st_gid = getgid();
	buf->st_rdev = 0;
	buf->st_size = ip->size;
	buf->st_blksize = 512;
	buf->st_blocks = (ip->size+511)&~511;
	buf->st_atime = fd->rf->hdr.time;
	buf->st_mtime = fd->rf->hdr.time;
	buf->st_ctime = fd->rf->hdr.time;
	return(0);
}

int
res_lstat(const char *file_name, struct stat *buf)
{
	return(res_rstat(NULL, file_name, buf));
}

off_t
res_lseek(struct RES_FD *fd, off_t offset, int whence)
{
	if (fd->rf->have_res) return(lseek(fd->a1.fd, offset, whence));
	switch (whence) {
	case SEEK_SET:
		if (offset < 0 || offset > fd->size)
			return(-1);
		fd->offset = offset;
		return(0);
	case SEEK_CUR:
		fd->offset += offset;
		if (fd->offset < 0 || fd->offset > fd->size) {
			fd->offset = 0;
			return(-1);
		}
		return(0);
	case SEEK_END:
		if (offset == 0) {
			fd->offset = fd->size;
			return(0);
		}
		// fall thru
	default:
		errno = EINVAL;
		return(-1);
	}
}

RES_FILE *
res_fopen(struct RES_ENV *xrf, const char *name, const char *mode)
{
	struct RES_FILE *f;
	struct RES_ENV *rf;
	struct RES_FD *fd;
	char *cp;

	if (xrf == NULL) {
		rf = res_open_env(NULL);
		if (rf == NULL) {
			errno = ENOENT;
			return(NULL);
		}
	} else {
		rf = xrf;
	}
	if (rf->have_res) {
		FILE *ff;

		cp = tmp_path(name);
		if (cp == NULL)
			return(NULL);
		ff = fopen(cp, mode);
		free(cp);
		if (ff) {
			fd = (struct RES_FD *)malloc(sizeof(*fd));
			if (fd) {
				fd->rf = rf;
				fd->a1.fd = 0;
				fd->a2.f = ff;
				fd->a3.dp = NULL;
				f = (struct RES_FILE *)malloc(sizeof(*f));
				if (f) {
					f->err = 0;
					f->fd = fd;
				} else {
					free(fd);
					fclose(ff);
					if (xrf == NULL)
						res_close_env(rf);
				}
			} else {
				fclose(ff);
				if (xrf == NULL)
					res_close_env(rf);
			}
		} else {
			if (xrf == NULL)
				res_close_env(rf);
			f = NULL;
		}
		return(f);
	}
	fd = res_open(rf, name, O_RDONLY); 
	if (xrf == NULL)
		res_close_env(rf);
	if (fd == NULL) {
		return(NULL);
	}
	f = (RES_FILE *)malloc(sizeof(&f));
	if (f == NULL) {
		res_close(fd);
		return(NULL);
	}
	f->fd = fd;
	f->err = 0;
	return(f);
}

int
res_fclose(RES_FILE *stream)
{
	if (stream->fd->rf->have_res) 
		fclose(stream->fd->a2.f);
	res_close(stream->fd);
	free(stream);
	return(0);
}

size_t
res_fread( void *ptr, size_t size, size_t nmemb, RES_FILE *stream)
{
	res_u64 s, r;

	if (stream->fd->rf->have_res) return(fread(ptr, size, nmemb, stream->fd->a2.f));
	s = size*nmemb;
	r = res_read(stream->fd, ptr, s);
	if (r == 0) {
		return(0);
	}
	if (r < size) {
		stream->err = errno;
		return(-1);
	}
	return(r/size);
}

void
res_clearerr(RES_FILE *stream)
{
	if (stream->fd->rf->have_res) {
		clearerr(stream->fd->a2.f);
		return;
	}
	stream->err = 0;
}

int 
res_feof(RES_FILE *stream)
{
	if (stream->fd->rf->have_res) return(feof(stream->fd->a2.f));
	return(stream->fd->offset == stream->fd->size);
}

int 
res_ferror(RES_FILE *stream)
{
	if (stream->fd->rf->have_res) return(ferror(stream->fd->a2.f));
	return(stream->err);
}

struct RES_FD *
res_fileno(RES_FILE *stream)
{
	return(stream->fd);
}

int
res_fgetc(RES_FILE *stream)
{
	unsigned char c;
	int i;

	if (stream->fd->rf->have_res) return(fgetc(stream->fd->a2.f));
	i = res_fread(&c, 1, 1, stream);
	if (i == 1)
		return(c);
	return(EOF);
}

char *
res_fgets(char *s, int size, RES_FILE *stream)
{
	char *cp = s;
	int i, j;
	char c;

	if (stream->fd->rf->have_res) return(fgets(s, size, stream->fd->a2.f));
	for (i = 0; i < size; i++) {
		j = res_fread(&c, 1, 1, stream);
		if (j <= 0) {
			if (i == 0)
				return(NULL);
			*cp = 0;
			return(s);
		}
		*cp++ = c;
		if (c == '\n') {
			if (i == (size-1)) 
				cp--;
			*cp = 0;
			return(s);
		} 
	}
	cp[-1] = 0;
	return(s);
}

int 
res_getc(RES_FILE *stream)
{
	int i;
	unsigned char c;

	if (stream->fd->rf->have_res) return(getc(stream->fd->a2.f));
	i = res_fread(&c, 1, 1, stream);
	if (i == 1)
		return(c);
	return(EOF);
}

int
res_ungetc(int c, RES_FILE *stream)
{
	if (stream->fd->rf->have_res) return(ungetc(c, stream->fd->a2.f));
	if (stream->fd->offset)
		stream->fd->offset--;
	return(0);
}

int
res_fseek(RES_FILE *stream, long offset, int whence)
{
	if (stream->fd->rf->have_res) return(fseek(stream->fd->a2.f, offset, whence));
	return(res_lseek(stream->fd, offset, whence));
}

long
res_ftell(RES_FILE *stream)
{
	if (stream->fd->rf->have_res) return(ftell(stream->fd->a2.f));
	return(stream->fd->offset);
}

void
res_rewind(RES_FILE *stream)
{
	if (stream->fd->rf->have_res) {
		rewind(stream->fd->a2.f);
		return;
	}
	stream->fd->offset = 0;
	stream->err = 0;
}

int
res_fgetpos(RES_FILE *stream, fpos_t *pos)
{
	if (stream->fd->rf->have_res) return(fgetpos(stream->fd->a2.f, pos));
	*pos = stream->fd->offset;
	return(0);
}

int
res_fsetpos(RES_FILE *stream, fpos_t *pos)
{
	if (stream->fd->rf->have_res) return(fsetpos(stream->fd->a2.f, pos));
	stream->fd->offset = *pos;
	if (stream->fd->offset < 0 || stream->fd->offset > stream->fd->size) {
		stream->fd->offset = 0;
		return(-1);
	}
	return(0);
}



RES_DIR *
res_opendir(struct RES_ENV *xrf, const char *name)
{
	struct RES_ENV *rf;
	struct RES_DIR *dp;
	struct RES_FD *fd;

	if (xrf == NULL) {
		rf = res_open_env(NULL);
		if (rf == NULL) {
			errno = ENOENT;
			return(NULL);
		}
	} else {
		rf = xrf;
	}
	if (rf->have_res) {
		DIR *d;

		name = tmp_path(name);
		if (name == NULL)
			return(NULL);
		d = opendir(name);
		free(name);
		if (d) {
			fd = (struct RES_FD *)malloc(sizeof(*fd));
			if (fd) { 
				fd->rf= rf;
				fd->a1.fd = 0;
				fd->a2.f = NULL;
				fd->a3.dp = d;
				dp = (struct RES_DIR *)malloc(sizeof(*dp));
				if (dp) {
					dp->fd = fd;
				} else {
					free(fd);
					if (xrf == NULL)
						res_close_env(rf);
					closedir(d);
					
				}
			} else {
				if (xrf == NULL)
					res_close_env(rf);
				closedir(d);
				dp = NULL;
			}
		} else {
			if (xrf == NULL)
				res_close_env(rf);
			dp = NULL;
		}
		return(dp);
	}
	fd = get_fd(rf, name, 0, 1);
	if (xrf == NULL)
		res_close_env(rf);
	if (fd == NULL)
		return(NULL);
	dp = (RES_DIR *) malloc(sizeof(*dp));
	if (dp == NULL) {
		res_close(fd);
		return(NULL);
	}
	dp->fd = fd;
	dp->offset = -2;				// mark '.' and next
	dp->base = fd->base;
	return(dp);
}

int
res_closedir(RES_DIR *dp)
{
	if (dp->fd->rf->have_res) 
		closedir(dp->fd->a3.dp);
	res_close(dp->fd);	
	free(dp);
	return(0);
}

struct dirent *
res_readdir(RES_DIR *dp)
{	
	struct res_inode *ip;

	if (dp->fd->rf->have_res) return(readdir(dp->fd->a3.dp));
	if (dp->offset == 0)
		return(NULL);
	dp->d.d_off = dp->offset; 
	if (dp->offset == -2) {
		strcpy(dp->d.d_name, ".");
		dp->d.d_reclen = 1;
		dp->d.d_ino = dp->fd->a1.inode;
		dp->offset++;
	} else
	if (dp->offset == -1) {
		strcpy(dp->d.d_name, "..");
		dp->d.d_reclen = 2;
		dp->d.d_ino = dp->fd->a2.pinode;
		dp->offset = dp->base;
	} else {
		ip = rd_ip(dp->fd->rf, dp->offset);
		if (ip->next == 0) {
			dp->offset = 0;
		} else {
			dp->offset += ip->next;
		}
		strncpy(dp->d.d_name, ip->name, sizeof(dp->d.d_name));
		dp->d.d_name[sizeof(dp->d.d_name)-1] = 0;
		dp->d.d_reclen = strlen(dp->d.d_name);
		dp->d.d_ino = ip->inode;
	}
	return(&dp->d);
}

off_t
res_telldir(RES_DIR *dp)
{
	if (dp->fd->rf->have_res) return(telldir(dp->fd->a3.dp));
	return(dp->offset);
}

void
res_seekdir(RES_DIR *dp, off_t offset)
{
	if (dp->fd->rf->have_res) {
		seekdir(dp->fd->a3.dp, offset);
		return;
	}
	dp->offset = offset;
	return;
}


void
res_rewinddir(RES_DIR *dp)
{
	if (dp->fd->rf->have_res) {
		rewinddir(dp->fd->a3.dp);
		return;
	}	
	dp->offset = -2;
}
