/* 
   Copyright (c) 2001 Paul Campbell - paul@taniwha.com
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.
 
   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.
 
   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/
                           
#ifndef __RESOURCES_H
#define __RESOURCES_H 1
#include <stdio.h> 
#include <sys/types.h>
#include <unistd.h> 
#include <dirent.h> 
#include <sys/stat.h>
#include <time.h>

//
//	resource section format
//
//	the resource section in an ELF file is
//	a non-loadable section, it is called "resource" and is intended to
//	be treated as a read-only 'file system' - it is also intended to
//	potentially be available at /procs/NNNN/resource/... - as such
//	it has a hierarchy, 'inodes' with standard attributes
//	and of course data in the 'files'
//
//
//	file formats are 'native endian' sized - all structures are
//	4-byte aligned
//
//


//
// file structures
//

typedef unsigned long long res_u64;
typedef          long long res_i64;

struct res_inode {
	res_u64		offset;				// offset from start of section to data
	res_u64		parent;				// offset from start of section to parent inode
	res_u64		size;				// length of section
	unsigned long	next;				// offset to next entry (0 at end)
	int		nlinks;				// number of links (includes phantom . and .. links )
	mode_t		mode;				// fs mode
	ino_t		inode;				// ours
	ino_t		pinode;				// parent's
	char		name[0];			// actual name
};

#define RES_VERSION_1_0	"RES_RV10"
struct res_hdr {
	char			version[8];		// the version of this file structure
	time_t			time;			// creation time
	struct res_inode	root;			// root inode
	char			comment[0];		// creation comment/copyright
};




//
//	structure describing an open resource file
//
typedef struct RES_ENV {
	res_u64 location;		// current location
	struct res_hdr hdr;		// header we got from the file
	FILE *f;			// open code file
	int ref_count;			// number of reasons we have the code file open
	unsigned char have_res;		// true if we have a /proc/self/resource
	unsigned long total_base;	// offset to start of our section within the file
	unsigned long total_size;	// size of the section
	unsigned char ip[NAME_MAX+2+sizeof(struct res_inode)];	// temp for an inode plus text
} RES_ENV;

typedef struct RES_FD {
	union {
		ino_t		inode;	// info about an open file
		int		fd;	// fd for the kernel case
	}		a1;
	union {
		ino_t		pinode;	// my parent 
		FILE		*f;	// FILE for the kernel case
	}		a2;
	union {
		mode_t		mode;	// mode
		DIR 		*dp;	// DIR for the kernel case
	}		a3;
	res_u64		base;		// base offset within section of data
	res_u64		offset;		// current offset
	res_u64		parent;		// address of parent's inode
	res_u64		addr;		// address of my inode
	res_u64		size;		// inode's size
	struct RES_ENV	*rf;		// the resource file it's in
} RES_FD;

typedef struct RES_FILE {
	struct RES_FD	*fd;		// file also has error state
	int		err;
} RES_FILE;

typedef struct RES_DIR {
	res_u64		offset;		// offset -2/-1 are ./.., 0 is eof, rest are &next_inode
	res_u64		base;		// base of list of inodes
	struct RES_FD	*fd;		// the current open fd
	struct dirent	d;
} RES_DIR;

//	files to open resource files
//
RES_ENV *res_open_env(const char *name);	
void res_close_env(RES_ENV *);	
//
//	these routines are analogs of the traditional
//	low level file routines
//
RES_FD 	*res_open(RES_ENV*, const char *pathname, int flags);
int 	res_close(RES_FD *fd);
ssize_t res_read(RES_FD *fd, void *buf, size_t count);
int 	res_rstat(RES_ENV*, const char *file_name, struct stat *buf);
int 	res_stat(const char *file_name, struct stat *buf);
int 	res_fstat(RES_FD *filedes, struct stat *buf);
int 	res_lstat(const char *file_name, struct stat *buf);
off_t 	res_lseek(RES_FD *fildes, off_t offset, int whence);

//
//	these are the stdout analogs (inputs only)
//
RES_FILE *res_fopen(RES_ENV *, const char *name, const char *mode);
int 	res_fclose(RES_FILE *stream);
size_t 	res_fread( void *ptr, size_t size, size_t nmemb, RES_FILE *stream);
void 	res_clearerr(RES_FILE *stream);
int 	res_feof(RES_FILE *stream);
int 	res_ferror(RES_FILE *stream);
int 	res_fgetc(RES_FILE *stream);
char 	*res_fgets(char *s, int size, RES_FILE *stream);
int 	res_getc(RES_FILE *stream);
int 	res_ungetc(int c, RES_FILE *stream); 
int 	res_fseek(RES_FILE *stream, long offset, int whence);
long 	res_ftell(RES_FILE *stream);
void 	res_rewind(RES_FILE *stream);
int 	res_fgetpos(RES_FILE *stream, fpos_t *pos);
int 	res_fsetpos(RES_FILE *stream, fpos_t *pos);  

//
//	these are the directory access routines
//
RES_DIR *res_opendir(RES_ENV *, const char *name);
int 	res_closedir(RES_DIR *dp);
struct dirent *res_readdir(RES_DIR *dp);
off_t 	res_telldir(RES_DIR *dp);
void 	res_seekdir(RES_DIR *dp, off_t offset);
void 	res_rewinddir(RES_DIR *dp);
#endif
